/* 
 * OpenXML (docx) to text conversion.
 * Steve Simon,   Oct 2012
 */
#include <u.h>
#include <libc.h>
#include <bio.h>
#include <xml.h>
#include <ctype.h>

static int para(Biobuf *bp, Elem *ep);

#pragma	   varargck    type  "M"   char *

static int Keepblanks;

int
Mfmt(Fmt *f)
{
	int n;
	char *p;
	Rune r;

	n = 0;
	p = va_arg(f->args, char *);
	while(*p){
		p += chartorune(&r, p);

		if(r == 0xf0b7)				/* Microsoft seem to use this (private code page) as a bullet */
			r = 0x2022;

		if(fmtprint(f, "%C", r) < 0)
			return -1;
		if(! isspacerune(r))
			n++;
	}
	return n;
}

static int
object(Biobuf *bp, Elem *ep)
{
	int n;
	char *v, *p;

	n = 0;
	for(; ep; ep = ep->next)
		if(strcmp(ep->name, "o:OLEObject") == 0)
			if((v = xml_value(ep, "ProgID")) != nil){
				/* trim cruft */
				if((p = strrchr(v, '.')) != nil && atoi(p+1) != 0)
					*p = 0;
				n += Bprint(bp, "%M", v);
			}

	return n;
}
	
static int
run(Biobuf *bp, Elem *ep)
{
	int n;

	n = 0;
	for(; ep; ep = ep->next){
		if(strcmp(ep->name, "w:drawing") == 0)
			n += Bprint(bp, "[drawing]");

		if(strcmp(ep->name, "w:pict") == 0)
			Bprint(bp, "[picture]");

		if(strcmp(ep->name, "w:object") == 0 && ep->child){
			n += Bprint(bp, "[object: ");
			n += object(bp, ep->child);
			n += Bprint(bp, "]");
		}

		if(strcmp(ep->name, "w:t") == 0 && ep->pcdata)
			n += Bprint(bp, "%M", ep->pcdata);

		if(strcmp(ep->name, "w:tab") == 0)
			n += Bprint(bp, "\t");
	}
	return n;
}


static int
cell(Biobuf *bp, Elem *ep)
{
	int first, n;

	n = 0;
	first = 1;
	for(; ep; ep = ep->next)
		if(strcmp(ep->name, "w:p") == 0 && ep->child){
			if(! first)
				Bprint(bp, " ");
			first = 0;
			n += para(bp, ep->child);
		}
	return n;
}

static int
row(Biobuf *bp, Elem *ep)
{
	int n;

	n = 0;
	for(; ep; ep = ep->next){
		if(strcmp(ep->name, "w:tc") == 0){
			Bprint(bp, "\t");
			n += cell(bp, ep->child);
		}
	}
	if(n || Keepblanks)
		Bprint(bp, "\n");
	return 0;
}

static int
table(Biobuf *bp, Elem *ep)
{
	int n;

	n = 0;
	for(; ep; ep = ep->next)
		if(strcmp(ep->name, "w:tr") == 0 && ep->child)
			n += row(bp, ep->child);
	return n;
}

static int
paraprops(Biobuf *bp, Elem *ep)
{
	int n;

	n = 0;
	for(; ep; ep = ep->next)
		if(strcmp(ep->name, "w:numPr") == 0 && ep->child)
			n += Bprint(bp, "    • ");
	return n;
}
	
static int
para(Biobuf *bp, Elem *ep)
{
	int n;

	n = 0;
	for(; ep; ep = ep->next){
		if(strcmp(ep->name, "w:pPr") == 0 && ep->child)
			n += paraprops(bp, ep->child);

		if(strcmp(ep->name, "w:r") == 0 && ep->child)
			n += run(bp, ep->child);

		if(strcmp(ep->name, "w:hyperlink") == 0 && ep->child)
			n += para(bp, ep->child);

		/* autogenerated fields, e.g. auto numbered Figures */
		if(strcmp(ep->name, "w:fldSimple") == 0 && ep->child)
			n += para(bp, ep->child);
	}
	return n;
}

static int
body(Biobuf *bp, Elem *ep)
{
	int n;

	n = 0;
	for(; ep; ep = ep->next){
		if(strcmp(ep->name, "w:p") == 0 && ep->child){
			n += para(bp, ep->child);
			if(n || Keepblanks){
				Bprint(bp, "\n");
				n = 0;
			}
			Bprint(bp, "\n");
		}
		if(strcmp(ep->name, "w:tbl") == 0 && ep->child){
			if(n || Keepblanks){
				Bprint(bp, "\n");
				n = 0;
			}
			table(bp, ep->child);
		}
	}
	return n;
}
static void
usage(void)
{
	fprint(2, "usage: %s [-b] [file.xml]\n", argv0);
	fprint(2, "           -b      keep all blank lines\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	Xml *xp;
	Elem *ep;
	char *err;
	Biobuf *bp, inp, bout;

	err = nil;
	ARGBEGIN{
	case 'b':
		Keepblanks++;
		break;
	case 'd':
		xml_debug++;
		break;
	default:
		usage();
	}ARGEND;

	if(argc == 0){
		Binit(&inp, 0, OREAD);
		if((xp = xml_Bparse(&inp, 8192, 0)) == nil)
			sysfatal("stdin: %r\n");
		Bterm(&inp);
	}
	else{
		if((bp = Bopen(argv[0], OREAD)) == nil)
			sysfatal("%s cannot open\n", argv[0]);
		if((xp = xml_Bparse(bp, 8192, 0)) == nil)
			sysfatal("%s: %r\n", argv[0]);
		Bterm(bp);
	}

	fmtinstall('M', Mfmt);

	Binit(&bout, 1, OWRITE);
	if((ep = xml_look(xp->root, "/w:document/w:body", nil, nil)) == nil || ep->child == nil)
		err = "bad xml format";
	body(&bout, ep->child);
	Bterm(&bout);

	xml_free(xp);
	exits(err);
}

